from collections import defaultdict
import json
import pathlib
from typing import Callable, DefaultDict, Dict, Iterable, List, Literal, Optional, Union
from crtl.components.crtl_el import get_setters_and_getters
from crtl.lkb.lkb import Concept, LexicalKnowledgeBase
from spacy.tokens import Doc, Span, SpanGroup
from spacy.language import Language
from spacy.util import registry
from spacy.pipeline import Pipe
from spacy.training import Example


class DiscrepancyRectifier(Pipe):
    def __init__(
        self, ent_or_span: Literal["ent", "span"], span_key: Optional[str] = None
    ):
        self.span_getter, self.span_setter = get_setters_and_getters(
            ent_or_span, span_key
        )

    def initialize(
        self,
        get_examples: Callable[[], Iterable[Example]],
        lkb_getter: Callable[[], LexicalKnowledgeBase],
        cui_to_cer_label_mapping: Dict[str, str],  # cui-to-cer_label mapping
        *,
        nlp=None,
    ):
        lkb = lkb_getter()
        self.cui_to_cer_label_mapping = cui_to_cer_label_mapping
        self.low_to_high_map = self._generate_low_to_high_mapping(lkb)

    def __call__(self, doc: Doc) -> Doc:
        rectified_spans = []
        for span in self.span_getter(doc):
            if span.kb_id_ in self.low_to_high_map:
                cer_labels = [
                    self.cui_to_cer_label_mapping[kb_id]
                    for kb_id in self.low_to_high_map[span.kb_id_]
                ]
                if not span.label_ in cer_labels:
                    span.kb_id_ = "NIL"
            else:
                span.kb_id_ = "NIL"

            rectified_spans.append(span)
        self.span_setter(doc, rectified_spans)
        return doc

    def to_disk(self, path, exclude=tuple()):
        if not path.exists():
            path.mkdir()
        with open(path / "high_level_kb_id_to_name.json", "w+") as f:
            json.dump(self.cui_to_cer_label_mapping, f)
        with open(path / "low_to_high_map.json", "w+") as f:
            json.dump(self.low_to_high_map, f)

    def from_disk(self, path, exclude=tuple()):
        with open(path / "high_level_kb_id_to_name.json") as f:
            self.cui_to_cer_label_mapping = json.load(f)
        with open(path / "low_to_high_map.json") as f:
            self.low_to_high_map = json.load(f)

    def _generate_low_to_high_mapping(
        self, lkb: LexicalKnowledgeBase
    ) -> DefaultDict[Union[str, int], List[Union[str, int]]]:
        # Function to collect the nodes in a breadth-first traversal
        low_to_high_map = defaultdict(lambda: [])
        for high_level_concept_id in self.cui_to_cer_label_mapping.keys():
            high_level_concept = lkb.cui_to_concept[high_level_concept_id]
            low_to_high_map[high_level_concept.concept_id].append(
                high_level_concept.concept_id
            )
            descendants = self._get_descendants(high_level_concept, lkb)
            for descendant in descendants:
                low_to_high_map[descendant.concept_id].append(
                    high_level_concept.concept_id
                )
        return low_to_high_map

    @classmethod
    def _get_descendants(
        cls, concept: Concept, lkb: LexicalKnowledgeBase
    ) -> List[Concept]:
        descendants = []
        visited = set()
        queue = []
        queue.append(concept)
        visited.add(concept.concept_id)
        while queue:
            concept = queue.pop(0)
            for concept_relation in concept.related_concepts:
                child_concept = lkb.cui_to_concept[concept_relation.concept_id]
                if child_concept.concept_id not in visited:
                    if concept_relation.relation in ["inverse_isa", "has_tradename"]:
                        queue.append(child_concept)
                        descendants.append(child_concept)
                    visited.add(concept.concept_id)
        return descendants


@Language.factory(
    name="trend_discrepancy_rectifier",
    default_config={"ent_or_span": "ent", "span_key": None},
)
def make_trend_discrepancy_rectifier(
    nlp: Language,
    name: str,
    ent_or_span: Literal["ent", "span"],
    span_key: Optional[str],
):
    return DiscrepancyRectifier(ent_or_span, span_key)


@registry.readers("crtl.read_concept_ids_from_json")
def make_high_level_concept_id_reader(
    path: Union[str, pathlib.Path]
) -> Union[List[str], List[int]]:
    with open(path) as f:
        return json.load(f)
